#!/usr/bin/perl
#
# Wrapper for Markdown
# Adds Frogscript extensions to Markdown

use FrogUtils::utility;		# Frogscript utility subs.
use FrogUtils::MarkdownW;	# Markdown.pl 1.01 converted to be a real package.
use FrogUtils::SmartyPantsW;	# SmartyPants.pl 1.51 converted to be a real package.
use Text::Wrap;			# Hard wraps text
use File::Path;			# GoDaddy uses a really old version of this
						# make_path() isn't there.
						# Need to use the "legacy" mkpath();
						# http://perldoc.perl.org/File/Path.html
use Digest::MD5 qw(md5_hex);	# For escaping characters
use strict;				# Uninitialized variables are a fatal error.

my $Version = "MarkdownWrapper 0.40  2014-03-26";


# Set reasonable default values.  All of these can be overridden with 
# variables at the beginning of the document.
my $UseSmartyPants = "";	# Use SmartyPants to deal with "smart" quotes, emdashes, etc.?
my $ForceImageCreation = ""; # Always create output image? (Even if it already exists?)
my $OnlyPrintConfigs = "";	# Print the values of config variables and then stop.
							# For debugging bad problems.
my $Debug = "";			# Write debugging info to the output file
my $ShowVars = "";		# Also write %TampVars hash template variables

my $MakeTextFile = "1";	# Convert the source file into a word-wrapped .text file?
my $MakeHTMLFile = "1";	# Write the HTML file?  ("" for debugging path logic...)
my $CreateDirs = "1";	# Create output directories?  (See above.)

my $MakeTheseJPEGs = "png|tif|tiff";	# Make images files with these file extension be JPEGs

# Not used (for now.)
# Character formatting.  %b%Text to be bolded%/b%  etc.  Works, shrug.
my $AlternateFontBegin = "<font face = Arial, Helvetica, \"san-serif\">";
my $AlternateFontEnd = "</font>";	# Will eventually be replaced with CSS.
$AlternateFontBegin = "<span style=\"font-family : arial, geneva, helvetica, sans-serif;\">";
$AlternateFontEnd = "</span>";

my $MaxDirLevels = 4;	# Maximum number of directory levels. (Runaway $JustDir...) 
my $DDebug = "1";	# Enable whatever quick hack print statements we are working on.

my $NLDir = "FLinks/";	# Where named links will go.

print "\n$Version\n";

my ($s, $s1, $arg, $pwd);
my $ErrLog = "";
my $fatal = 0;
$arg = shift;
$arg =~ s|\\|/|g;

my $Config = shift;
my $DebugConfig = shift;
unless ($Config) { 
	$ErrLog .=  "Usage: InputFile ConfigFile\n";
	$ErrLog .= "You didn't supply a configuration file.\n";
	$fatal++;
	$Config = "";
	$DebugConfig = "Debug";	# Let's just force displaying the ErrLog and then stopping.
}
my $ext = "";
if ($Config =~ m/debug/i) {
	$Config = "";
	$DebugConfig = "Debug";	# Let's just force displaying the ErrLog and then stopping.
	#$fatal++;
} else { 
	$ext = return_fext($Config);
	$Config = return_file(return_fname($Config))  . ".cfg"; 	# Strip off extension and put it back again
}
my $ConfigDir = "etc/markcms/";	# Where the config files are found
								# (Appended to $HOME.)								
GetPwd($arg); 	# Deals with Windows (pwd = 'c:\Windows\System32' sometimes craziness

my $HomePath = "";
if ( $^O =~ /MSWin.*/i ) {
	$ErrLog .= "(Windows)\n";
	if ( $ENV{'FHOME'} ) {	# i.e., SET should contain 'FHOME=C:\HOME'
		$HomePath = $ENV{'FHOME'};
	} elsif ( $ENV{'HOME'} ) {
		$HomePath = $ENV{'HOME'};
	} elsif ( -d "C:/home" ) {	# Default install dir.  (Per instructions.)
		$HomePath = "C:/home";
	} else {
		$ErrLog .= "Environment variable 'FHOME' does not exist.\n";
		$fatal++;
	}
} else {
	$ErrLog .= "(Linux)\n";
	if ( $ENV{'FHOME'} ) {
		$HomePath = $ENV{'FHOME'};
	} elsif ( $ENV{'HOME'} ) {
		$HomePath = $ENV{'HOME'};
	} else {
		$ErrLog .= "Environment variable 'HOME' does not exist.  We need it.\n";
		$fatal++;
	}
}
$HomePath  =~ s|\\|/|g;
$HomePath = add_rslash($HomePath);
my $ConfigPath = $HomePath . $ConfigDir;
unless ( -d "$ConfigPath" ) { $ErrLog .= "Could not find config file directory '$ConfigPath'\n"; $fatal++ }
unless ( -f "$ConfigPath$Config" ) {
	#$ErrLog .= "Could not find config file '$ConfigPath$Config'\n"; 
	#$fatal++; 
}

my $CurrentDrive = ""; 
my $BaseInDir = "";	# Base directory where the authoring files are.
					# $pwd won't work, so it has to be specifically defined.
my $BaseOutDir = "";

# (Some of) these should be read from the config file.
if ( $^O =~ /MSWin.*/i ) {
	$CurrentDrive = GetDrive($pwd);
	$BaseInDir = add_rslash($pwd);
	$BaseOutDir = $HomePath . 'html/frog_html/';
	$BaseInDir = ChangeDrive($BaseInDir, $CurrentDrive);
	$BaseOutDir = ChangeDrive($BaseOutDir, $CurrentDrive);	
	$HomePath = ChangeDrive($HomePath, $CurrentDrive);
	$ConfigPath = ChangeDrive($ConfigPath, $CurrentDrive);
} else {
	$BaseInDir = add_rslash($pwd);
	$BaseOutDir = '~/html/frog_html/';
}

my $JustDir = $pwd;
# This breaks bad if $BaseInDir isn't accurate!
$JustDir =~ s/$BaseInDir//i;

my $InDir = add_rslash($BaseInDir . $JustDir);
my $OutDir = add_rslash($BaseOutDir . $JustDir);

my $FileName = return_file($arg);
$FileName =~ s|\\|/|g;
unless ( -f $FileName ) { $ErrLog .= "'$FileName' doesn't exist!\n"; $fatal++; }
my $BaseName = return_fname($FileName);
my $InFile = $FileName;
my $HtmlFile = $BaseName . ".html";
my $TextFile = $BaseName . ".text";
my $TmpFile1 = $BaseName . ".tmp1";	# Needed?
my $TmpFile2 = $BaseName . ".tmp2";	# Needed?
my $Path;
if ( $^O =~ /MSWin.*/i ) {
	$Path = add_rslash(return_path($pwd));	# This needs work.  More of the
											# 'Current directory == C:\Windows\System32' sometimes problem.
	unless ($Path) { add_rslash(return_path($pwd)) }
	#print "\$Path: '$Path' '$pwd'\n"; exit 0;
} else {
	$Path = add_rslash(return_path($pwd));
}
my $Ext = return_fext($FileName);

my $PageHeader = "";
my $PageFooter = "";
my $SideBarBegin = "<div class=\"rsidebar\">\n";
my $SideBarEnd = "</div>";
my $ImageBoxBegin = "";
my $ImageBoxEnd = "";
my $IDList = "";	# Holds any IDs (names) that were found.
my $NameLinks = "";
my %TempVars = ();	# Holds template variables
#my $RealOutFile = $OutDir . $TextFile;
$TempVars{"_version"} = $Version;

# For CSS sheets.
my %CSSVars = ();

my %SpanVars = ();	# For substituting <span =...>
#$SpanVars{"warning"} = "<span class=\"warning\">";
$SpanVars{"warning"} = "<span style=\"color: red; font-weight: bold;\">";
$SpanVars{"button"} = "<span class=\"button\">";

# Global variables that are (should be?) set with the config file.

my $MaxImageWidth = "640";	# Maximum width for images

my $IsMobile = "";			# Is it a mobile site? (Not used yet.)
my $FilesToBeCopied = "";	# Filled in by configuration file
# Enabling any of these affect runtime. (But are deprecated.)
my $TemplateVarsInBody = "1";	# Process template variables in the body?  
								#  (Affects run time speed...)
my $UseExtended = "1";		# Use extended features?

my %Parms = ();
my $ImageMagickConfig = "-unsharp 0x0.35+1+0";
my $CompressQuality = 80;
#$CompressQuality = "";

if ($Config) { 
	ReadConfigFile($ConfigPath . $Config);
}
if (-f "dirconfig.cfg") {
	ReadConfigFile("dirconfig.cfg");
}

$BaseInDir = FixLinuxPath($BaseInDir);
$BaseOutDir = FixLinuxPath($BaseOutDir);

$TempVars{"_self_text"} = $JustDir . $TextFile;
$TempVars{"_self_html"} = $JustDir . $HtmlFile;
$TempVars{"_justdir"} = $JustDir;
my $RelPath = "";
my $LevelCounter = 0;
my @counter = split(/\//, $JustDir);	# How many levels up are we from the base directory?
$TempVars{"_dirlevel"} = scalar @counter;
foreach (@counter) { $RelPath .= "../"; $LevelCounter++ }
unless ($RelPath) { $RelPath = "./" }
if ( $LevelCounter > $MaxDirLevels ) {
	#print $ErrLog;
	$ErrLog .= "\n\$JustDir '$JustDir' has too many levels.\n";
	$ErrLog .= "Maximum of $MaxDirLevels levels (\$MaxDirLevels)\n";
	$ErrLog .= "Check the paths in your .cfg file and try again.\n";
	$ErrLog .= "Especially 'BaseInDir'\n";
	$fatal++;
}

$TempVars{"_pathtohome"} = $RelPath;
#$TempVars{"_indir"} = $InDir;
#$TempVars{"_outdir"} = $OutDir;
#$TempVars{"_baseindir"} = $BaseInDir;
#$TempVars{"_baseoutdir"} = $BaseOutDir;
#$TempVars{"_"} = $;

my $DirMask = '(md$)|(mdf$)';	# '.md' will be deprecated.
unless ($Ext =~ m/$DirMask/i) { $ErrLog .= "Not a Frogdown '.md' or '.mdf' file.\n"; $fatal++ }
my $cfile = "";
if ( -f $InFile ) {
	$cfile = readfile($InFile);
	print "Read input file: '$InFile'\n";
	#readfile_byref($InFile, \$cfile)
	#unless ( readfile_byref($InFile, \$cfile) ) { print "$InFile not found\n"; exit -1 }
} else {
	$ErrLog .= "Document file '$InFile' not found\n"; $fatal++;
}
my $Text = GetHeadVars($cfile);	# Get any variables from the beginning of the file (and then delete them from the file)

if ( $DebugConfig =~ m/^Debug/i) { 
	$OnlyPrintConfigs = 1; # $fatal = 0;
	#$FilesToBeCopied = "(None: Didn't read any configuration file.)";	
}

if ($fatal && !($DebugConfig) && 0 ) {
	$ErrLog .= "\nToo many fatal errors ($fatal)\n";
	print $ErrLog;
	exit -1;
}

my $TempErrLog = $ErrLog; $ErrLog = "";

$ErrLog .= "\n";
$ErrLog .= "Basename:          '$BaseName'\n";
$ErrLog .= "Ext:               '$Ext'\n";
$ErrLog .= "Filename:          '$FileName'\n";
$ErrLog .= "htmlfile           '$HtmlFile'\n";
#$ErrLog .= "Tempfile           '$TmpFile'\n";
#$ErrLog .= "Textfile           '$TextFile'\n";
$ErrLog .= "Config             '$Config'\n";
$ErrLog .= "\n";
if ($CurrentDrive) { $ErrLog .= "CurrentDrive:      '$CurrentDrive'\n" }
$ErrLog .= "HomePath:          '$HomePath'\n";
$ErrLog .= "ConfigPath         '$ConfigPath'\n";
$ErrLog .= "Path (pwd):        '$Path'\n";
$ErrLog .= "real pwd:          '$pwd'\n";
$ErrLog .= "JustDir:           '$JustDir'\n";
$ErrLog .= "%%PathToHome%%:    '$RelPath'\n";
$ErrLog .= "BaseInDir:         '$BaseInDir'\n";
$ErrLog .= "BaseOutDir:        '$BaseOutDir'\n";
$ErrLog .= "InDir:             '$InDir\n";
$ErrLog .= "OutDir:            '$OutDir\n";
$ErrLog .= "FilestoBeCopied:   '$FilesToBeCopied'\n";
#$ErrLog .= "OnlyPrintConfigs:  '$OnlyPrintConfigs'\n";
#$ErrLog .= "\n";
#$ErrLog .= "MaxImageWidth:     '$MaxImageWidth'\n";

if ($TempErrLog) { $ErrLog .= "\n$TempErrLog" }

if ($OnlyPrintConfigs || $fatal) {
	print $ErrLog; 
	if ($fatal) {
		print "Too many fatal errors ($fatal)\n";
	} else {
		print "Debug mode--did not create any files.\n";
	}
	#print MakeVarLog();
	exit 0;
}
# At this point we are in the correct directory so we don't need errlog anymore.
my $DebugFile = $ErrLog;
$ErrLog = "";
my $WorkLog = "";

if ( 6 == 9) {
	# Process the input .md file with Markdown.pl
	my $ComSpec = "cat $InFile | Markdown.pl > $HtmlFile";
	print "\n$ComSpec\n";
	$ErrLog = `$ComSpec`;
	if ($ErrLog) { print "$ErrLog\n"; }

	# Copy the .md file to be a .text file (and maybe wordwrap it...?)
	$ComSpec = "cp -fp $InFile $TextFile"; 
	print "$ComSpec\n";
	$ErrLog = `$ComSpec`;
	if ($ErrLog) { print "$ErrLog\n"; }
}

if ($CreateDirs) {
	# Create output directories if they don't already exist.
	unless ( -d $OutDir ) {
		print "Making $OutDir\n";
		mkpath($OutDir, 1);
	} else {
		#print "$OutDir exists.\n";
	}
} else {
	print "  (Not attempting to create $OutDir)\n";
}

if (6 == 9) {
	if ( -f $InFile ) {
		open (FILE, $InFile);
		$cfile = join('', <FILE>);	# Assumes that 'undef $/;' has *not* beeen done.
		close FILE;
		chomp $cfile;
	} else { print "'$InFile' not found.\n"; exit -1 }
	$cfile =~ s{\r\n}{\n}g; # DOS to Unix
	$cfile =~ s{\r}{\n}g; 	# Mac to Unix
}

my @FilesToBeCopied = split(/\n/, $FilesToBeCopied);
foreach (@FilesToBeCopied) {
	my $file = $_;
	my $RelFile =  $RelPath . $file;
	if ( -d $file ) {	# Umm, shouldn't this be '$InDir$file'?
		$ErrLog .= "Sorry, can't copy the directory '$file' yet.\n";
		# Doesn't work yet.
		#my $parm = "cp -Rfp \"$file/*\" \"$OutDir\" ";
		#$WorkLog .= "$parm\n";
		#my $result = `$parm`;
		#if ($result) { $ErrLog .= $result; }
	} else {
		my $ext = return_fext($file);
		my $basefile = return_fname($file);
		if ($ext =~ m/(csst)/i ) {
			#print "Reading '$RelFile'\n";
			if ( -f "$InDir$file" ) {
			#if ( -f "$RelFile" ) {
				my $s = readfile("$InDir$file");
				#my $s = readfile("$RelFile");
				ParseConfigCSSVars($s);	# Extract any CSS template variables
				$s = ParseCSSVars($s);	# Now substitute them.
				my $builtfile = $basefile . ".css";
				writefile_byref("$OutDir$builtfile", \$s);
				$WorkLog .= "Converted CSS template vars and wrote '$OutDir$builtfile'\n";
				#print $s;
				#print " '$basefile' '$ext'\n";
			} else {
				$ErrLog .= "File '$InDir$file' not found!\n";				
			}
		# } elsif ( -f "$BaseInDir$file" ) {
		} elsif ( -f "$InDir$file" ) {
			my $parm = "cp -fp \"$InDir$file\" \"$OutDir$file\" ";
			#my $parm = "cp -fp \"$BaseInDir$file\" \"$OutDir\" ";
			$WorkLog .= "$parm\n";
			my $result = `$parm`;
			if ($result) { $ErrLog .= $result; }
		} else {
			$ErrLog .= "\$InDir\$file     File '$InDir$file' not found!\n";
			#$ErrLog .= "\$BaseInDir\$file File '$BaseInDir$file' not found!\n";
		}
	}
}

my $x;
MakeTextFile($TextFile, $cfile);

# Cribbed from Markdown.pl
# '%' needs to be escaped *if* your Markdown source file
# contains anything that this script is looking for.
# i.e. If you write '%b%' it should be escaped as '\%b\%'
# Mainly comes up when describing what this script processes.
# (This script converts '%b%' to '<b>', etc.  So...)
# Not sure about '|' yet.
my %g_escape_table;
foreach my $char (split //, '|%=') {
	$g_escape_table{$char} = md5_hex($char);
}

$Text = PreFilter($Text);
$Text = Markdown($Text);
$Text = PostFilter($Text);
$Text = ParseSpanVars($Text);
if ($TemplateVarsInBody || 1 ) { $Text = ParseTempVars($Text) }

if ($Debug) {	
	my $ds = "";
	$ds .= "\n<!-- ";
	$ds .= "Created by $Version\n";
	$ds .= "\n\$Debugfile:\n" . $DebugFile;
	if ($ShowVars ) { $ds .= MakeVarLog(); }
	$ds .= "\n\$ErrLog:\n" . $ErrLog;
	$ds .= "\n\$Worklog:\n" . $WorkLog;
	$ds .= " -->\n\n";
	$Text = $ds . $Text;
} else {	# Well, we will at least write the $Version
	my $ds = "";
	$ds .= "\n<!-- ";
	$ds .= "Created by $Version\n";
	$ds .= " -->\n\n";
	$Text = $ds . $Text;
}

if ($UseSmartyPants) { 
	my $smartypants_attr = "1";  #  1 =>  "--" for em-dashes; no en-dash support								 
								 #  2 =>  "---" for em-dashes; "--" for en-dashes
								 #  3 =>  "--" for em-dashes; "---" for en-dashes
								 #  See SmartyPants docs for more configuration options.
	$Text = SmartyPants($Text, $smartypants_attr, undef); 
}

if ($IDList) { 
	unless ($IsMobile) {
		#print $IDList;
		$NameLinks = "\n<div>\n <ul class=\"NameLinks\">\n" . $NameLinks . " </ul>\n</div>\n";
		#print $NameLinks;
		$Text =~ s|<p>%%NameLinks%%</p>|$NameLinks|s;
	} else {
		$Text =~ s|%%NameLinks%%|<!-- Named links should go here -->|s;		# Sigh.  This does not work in Mobile jQuery
	}
	$NLDir = add_rslash($NLDir);
	unless ( -d $NLDir ) {
		print "Making $NLDir\n";
		mkpath($NLDir, 1);
	} else {
		#print "$OutDir exists.\n";
	}
	if ( 6 == 9) {	# Not sure if writing this out would be useful for anything.
		my $RealOutFile = $NLDir . "links_" . $BaseName . ".html";
		$WorkLog .= "Created Named Links file: $RealOutFile\n";
		writefile_byref ($RealOutFile, \$NameLinks);
	}
	my $IHead = "### Start of Variables ###\n";
	$IHead .= "# Created by $Version\n";
	$IHead .= "\n";
	$IHead .= "PageTitle = " . $TempVars{"pagetitle"} . "\n";
	$IHead .= "PageID = " . $TempVars{"pageid"} . "\n";
	$IHead .= "Filename = $HtmlFile\n";
	$IHead .= "JustDir = $JustDir\n";
	$IHead .= "RelPathToHome = $RelPath\n";
	$IHead .= "Fields = ID|HeadingLevel|Description|Title\n";
	$IHead .= "\n";
	$IHead .= "### End of Variables ###\n";
	$IDList = $IHead . $IDList;
	my $RealOutFile = $NLDir . $BaseName . ".txt";
	$WorkLog .= "Created Named Links file: $RealOutFile\n";
	writefile_byref ($RealOutFile, \$IDList);
}

while ($Text =~ s|\n\n\n|\n\n|g) {}

my $Head = addhead();
my $Tail = addtail();

my $RealOutFile = $OutDir . $HtmlFile;
$WorkLog .= "Created HTML file: $RealOutFile\n";
if ($MakeHTMLFile) {
	open (OUT, ">$RealOutFile") || die "Cannot open $RealOutFile";
	print OUT $Head;

	print OUT $Text;
	print OUT $Tail;
	print OUT "\n";
	close OUT;
} else {
	print "  (Not really creating $RealOutFile.)\n";
}

if ($WorkLog) { print "Worklog:\n$WorkLog\n"; }
if ($ErrLog) { print "Errlog:\n$ErrLog"; }

exit 0;

# Get current working directory.  Which is surprisingly complex. On Windows.
sub GetPwd {
	my $arg = shift;	# Caller *must* pass a filespec that exists in the real cwd.

	$pwd = `pwd`;
	$pwd = prep_path($pwd);


	if ( $^O =~ /MSWin.*/i ) {
		if ($pwd =~ m|c:/Windows/System32|i ) {
			print "We are starting out with '$pwd' as the cwd.\n";
			print "Current working directory is '$pwd'!\n";
			if (-f $arg) {	# *Sometimes* the current directory (as seen by Perl scripts) is
							# 'C:\Windows\System32'.  Windows passes the filename fully qualified 
							# (with a path).  So we pick off the path and change to that directory
							# But sometimes Windows doesn't pass the filename fully qualified...
							# Ugly.


				$pwd = prep_path(return_path($arg));
				print "Fixed pwd:         '$pwd'\n";
				if ( -d $pwd) {		# This is the path we got from $arg
					#$ErrLog .= "$pwd exists\n";	# Yea.
					chdir($pwd);
				} else {
					print "Current directory '$pwd' does not exist\n"; 
					print "Usage: 'call MarkdownWrapper.pl $arg configfile'\n";
					print "(We derive the pwd from '$arg')\n";
					print $ErrLog;
					exit -1;
				}
				print "We are ending up with    '$pwd' as the cwd.\n";
			} else {
				print "\n";
				print "arg:               '$arg'\n";	
				print "pwd:               '$pwd'\n";
			}
		} else {
			# Well, nothing.  We already have a valid pwd from above.
			# We only need to deal with 'pwd == C:\Windows\System32' foolishness on Windows.
		} # if ($pwd =~ m|c:/Windows/System32|i ) {
	} # if ( $^O =~ /MSWin.*/i ) {

	return $pwd;
}

# Get a drive letter from a string, if it exists.
sub GetDrive {
	my $s = shift;
	
	my $d = "";
	$s = trim($s);
	if ($s =~ m|^([A-Z]:)|i) {
		$d = uc($1);
	}
	return $d;
}

sub ChangeDrive {
	my $string = shift;
	my $drive = shift;
	
	$drive = uc($drive);
	$drive =~ s/://;	# Remove any ':'
	$string = trim($string);
	#print "'$string' '$drive'\n";
	if ($string =~ m/^[A-Z]:/i) {
		$string =~ s/^[A-Z]/$drive/i
	}
	return $string;
}

# $OutDir is a global and is assumed to be already set.
sub MakeTextFile {
	my $TextFile = shift;
	my $cfile = shift;
	
	#print "\$MakeTextFile: '$MakeTextFile'\n";
	if ($MakeTextFile) {
		# Write a version of the input Markdown file that is hard wordrapped.
		# So it will look civilized if displayed in a browser.
		my $x = wrapit($cfile, 79);
		#$x = $cfile;
		my $RealOutFile = $OutDir . $TextFile;
		$WorkLog .= "Created text file: $RealOutFile\n";
		open (OUT, ">$RealOutFile") || die "Cannot open $RealOutFile";
		print OUT "$x\n";
		close OUT;
	}
}

sub MakeVarLog {
	my $s = "";

	$s .= "\n%TempVars:\n";
	my ($key, $value);
	foreach $key (sort(keys %TempVars)) {
		$s .= sprintf(" %-20s %s\n", $key . ":", $TempVars{$key});
	}
	$s .= "\n%SpanVars:\n";
	foreach $key (sort(keys %SpanVars)) {
		$s .= sprintf(" %-20s %s\n", $key . ":", $SpanVars{$key});
	}	
	$s .= "\n%CSSVars:\n";
	foreach $key (sort(keys %CSSVars)) {
		$s .= sprintf(" %-20s %s\n", $key . ":", $CSSVars{$key});
	}
	
	return $s;
}

# Frogdown changes we make *before* running the text through Markdown()	
sub PreFilter {
	my $s = shift;
	my ($needle, $key, $value, $ParaAttrib, $head, $tail);
	
	$s = _EncodeBackslashEscapes($s);

	# Include files
	# %%Include dogfood-include.inc%%
	$needle = "%%Include\s*(.*?)%%";
	while ( $s =~ m|$needle|s ) {
		my $File = $1;
		
		$File = trim($File);
		my $chunk ="";
		if ( -f $File ) { 
			$chunk = readfile($File);
			$WorkLog .= "Included file '$File'\n";
			MakeTextFile($File, $chunk);
		} else {
			$chunk = "\nInclude file '$File' not found.\n";
			$ErrLog .= "Include file '$File' not found\n";
		}
		$s =~ s|$needle|$chunk|s;	# Out with the old, in with the new.
	}

if (6 == 6) {
	#$s =~ s|%code%|<code>|sgi;		# code
	#$s =~ s|%/code%|</code>|sgi;
	$needle = "%code%(.*?)%/code%";
	while ( $s =~ m|$needle|s ) {
		my $chunk = "`" . $1 . "`";
		#$chunk = _UnescapeSpecialChars($chunk);
		$chunk = Markdown($chunk);
		$chunk =~ s|</?p>||g;
		#$chunk = _EncodeBackslashEscapes($chunk);
		#print "Code: '$chunk'\n";
		$s =~ s|$needle|$chunk|s;	# Out with the old, in with the new.
	}
}
	
	#$s =~ s|%blockquote%|\n<blockquote><p>|sgi;		# Block quote
	#$s =~ s|%/blockquote%|</p></blockquote>\n|sgi;

	$needle = "%blockxquotex%\n*(.*?)\n*%/blockxquotex%";	# We are now doing this in PostFilter()
	while ( $s =~ m|$needle|s ) {
		my $chunk = $1;
		$chunk = Markdown($chunk);
		$chunk = "\n<blockquote>\n" . $chunk . "</blockquote>\n";
		#print $chunk;
		$s =~ s|$needle|$chunk|s;	# Out with the old, in with the new.
	}

	# Converts to *Markdown* not HTML.
	$needle = "%verse(.*?)%\n?(.*?)%/verse%";
	while ( $s =~ m|$needle|s ) {
		# %sidebar  sb=rsidebar|t=More about the blog post%
		my $Parms = trim($1);
		my $chunk = $2;
		
		if ($Parms) {
			#print "%verse% parms: '$Parms'\n";
			Parmify($Parms);	# Fill %Parms hash with parameters
		} else {
			#print "No %sidebar% parms.\n";
		}
		
		my $Class = "verse";
		my $Style = "";
		my $StyleParams = "";
		my $StyleSep = "";
		if (exists($Parms{"c"})) { $StyleParams .= $StyleSep . "text-align: center"; $StyleSep = "; " }
		if (exists($Parms{"f"})) { 
			$Parms{"f"} =~ s/"//g;
			$StyleParams .= $StyleSep . "font-family: " . $Parms{"f"}; $StyleSep = "; " 
		}
		if (exists($Parms{"s"})) { $StyleParams .= $StyleSep . "font-size: " . $Parms{"s"}; $StyleSep = "; " }
		if (exists($Parms{"lm"})) { $StyleParams .= $StyleSep . "padding-left: " . $Parms{"lm"}; $StyleSep = "; " }	
		if (exists($Parms{"style"})) { $StyleParams .= $StyleSep . $Parms{"style"}; $StyleSep = "; " }
		if (exists($Parms{"class"})) { $Class = $Parms{"class"}; $StyleSep = "; " } 
		
		if ($StyleParams) { $Style = " style =\"" . $StyleParams . "\"" }
		
		my @lines = split(/\n/, $chunk);
		my $line="";
		my $Verse = "";
		$Verse .= "\n<div>\n";
		$Verse .= "<p$Style Class = \"$Class\">\n";
		#$Verse = "";
		foreach $line (@lines) {
			$line = rtrim($line);
#				$line =~ s|%b%|<b>|sgi;			# Bold
#				$line =~ s|%/b%|</b>|sgi;
#				$line =~ s|%i%|<i>|sgi;		# Italic
#				$line =~ s|%/i%|</i>|sgi;
			if ($line =~ s/^( +)//) {
				my $spaces = $1;
				#my $text = $2;
				$spaces =~ s/ /&nbsp;/g;
				#print "Leading spaces: '$spaces' '$text'\n";
				$line = $spaces . $line;
			}	# Make leading spaces non breaking.
			#$line = "'$line'";
			#$line .= "   \n";
			$line = Markify($line);
			$line .= "<br>\n";
			$Verse .= $line;
		}
		$Verse .= "</p>\n";
		$Verse .= "</div>\n\n";
		#print "Verse:\n$Verse";
		$s =~ s|$needle|$Verse|s;	# Out with the old, in with the new.
	}
	
	$needle = "%sidebar\s*?(.*?)%\n?(.*?)%/sidebar%";
	while ( $s =~ m|$needle|s ) {
		# %sidebar  sb=rsidebar|t=More about the blog post%
		my $Parms = trim($1);
		my $chunk = $2;

		$ParaAttrib = "sidebar";
		
		if ($Parms) {
			#print "%sidebar% parms: '$Parms'\n";
			Parmify($Parms);	# Fill %Parms hash with parameters
		} else {
			#print "No %sidebar% parms.\n";
		}		
		$head = "<!-- Start of sidebar -->\n";
		$head .= "<div class = \"rsidebar\">\n";	# Need config variable for start of sidebar
		$head = $SideBarBegin;	# Now is a config variable
		if (exists($Parms{"t"})) {
			$head =~ s/%%SidebarTitle%%/$Parms{"t"}/gi;
		}
		if (exists($Parms{"o"})) {
			my $o = $Parms{"o"};
			if ( $o =~ m/[lr]/i) {
				$head =~ s/%%SidebarOrientation%%/$o/gi;
			} else {
				$ErrLog .= "Sidebar: Invalid 'o' option: $o\n";
			}
		}
		if (exists($Parms{"sb"})) {
			$head =~ s/%%Sidebar%%/$Parms{"sb"}/gi;
		}
		if (exists($Parms{"pa"})) { $ParaAttrib = $Parms{"pa"}; }
		
		$head = ParseTempVars($head);	# Expand any '%%TempVar%% variables in the boilerplate
		$chunk = Markdown($chunk);	# Comes back as HTML.
		$chunk = Classify($chunk, "$ParaAttrib", "p");
		$tail = "</div>\n";	# Need config variable for end of sidebar
		$tail .= "<!-- End of sidebar -->";
		$tail = $SideBarEnd;	# Now is a config variable
		$tail = ParseTempVars($tail);
		$chunk = $head . $chunk . $tail;
		#print "$chunk\n";
		$s =~ s|$needle|$chunk|s;	# Out with the old, in with the new.
	}
	
	$needle = "%table (.*?)%\n*(.*?)%/table%";
	while ( $s =~ m|$needle|s ) {
		# %table  b=0|cs=0|cp=5%
		my $Parms = trim($1);
		my $chunk = $2;
		
		my $DivStyle="left";
		if ($Parms) { 	
			#print "table parms: '$Parms'\n";
			Parmify($Parms);
		}

		# Trying to left, right, or center the table.
		if (exists($Parms{"f"})) { $Parms{"ib"} =  $Parms{"f"} }
		if (exists($Parms{"ib"})) {	 # Not working yet.
			my $value = trim($Parms{"ib"});
			if ( $value =~ m/^(r)$/i) { 
				$DivStyle = "right"; 
			} elsif ( $value =~ m/^(c)$/i) { 
				$DivStyle = "center";
			}
		}	
		$DivStyle="left";	# Need to figure this out better.
		my $Border = 0;
		my $BGColor = "";
		my $CellSpacing = 0;
		my $CellPadding = 4;
		my $BoldFirstCell = 0;	# Bold the first cell in each row?
		my $BoldFirstRow = 0;	# Bold the first row in the table?
		my $Width = "";
		if (exists($Parms{"b"})) { $Border = $Parms{"b"} }
		if (exists($Parms{"bg"})) { $BGColor = " bgcolor = " . $Parms{"bg"} }
		if (exists($Parms{"w"})) { $Width = " width = " . $Parms{"w"} }
		if (exists($Parms{"cs"})) { $CellSpacing = $Parms{"cs"} }
		if (exists($Parms{"cp"})) { $CellPadding = $Parms{"cp"} }
		if (exists($Parms{"bfc"})) { $BoldFirstCell = 1 }
		if (exists($Parms{"bfr"})) { $BoldFirstRow = 1 }
		
		my @lines = split(/\n/, $chunk);
		my $line ="";
		my $rows = "";
		my ($tds, $tde);
		my $TDH = 0;
		#$rows ="<div style=\"text-align: $DivStyle\; padding-left: 1em\">\n";
		$rows .= "<div style=\"padding-left: 1em; margin-top: 1em; margin-bottom: 1em;\">\n";
		$rows .= " <table border=$Border cellspacing = $CellSpacing cellpadding=$CellPadding$BGColor$Width>\n";
		my $cell = "";
		my @cells;
		my $CCnt = 0;
		my $RCnt = 0;
		foreach $line (@lines) {
			$tds = "<td>"; $tde = "</td>";
			$line = trim($line);
			if ( $line =~ m/^!/ ) {
				$tds = "<th>"; $tde = "</th>";
				$TDH = 1;
				$line =~ s/^!//;
			} else {
				$tds = "<td>"; $tde = "</td>";
				$TDH = 0;
			}
			unless ($line) { $rows .= "<!-- blank line -->\n"; next }	# Skip blank lines
			$line = Markify($line);
			@cells = split(/\|/, $line);
			$rows .="  <tr valign=\"top\">\n";
			$CCnt = 0;
			foreach $cell (@cells) {
				$cell = trim($cell);
				if ( $BoldFirstCell && !($CCnt) && !($TDH) ) { 
					$cell = "<b>$cell</b>";
				} elsif ( $BoldFirstRow && !($RCnt) && !($TDH) ) { 
					$cell = "<b>$cell</b>";
				}
				$rows .= "   $tds$cell$tde\n";
				$CCnt++;
			}
			$rows .= "  </tr>\n";
			$RCnt++;
		}
		$rows .= " </table></p>\n";
		#$rows .= "</div> <!-- End of table -->\n";
		$rows .= "</div>\n\n";
		#$chunk = "<pre>\n$rows\n</pre>\n";
		$s =~ s|$needle|$rows|s;	# Out with the old, in with the new.
	}
	$s =~ s|</table></p>|</table>|sg;

	$needle = "%id (.*?)%\n(#{1,5})(.*?)\n";
	while ( $s =~ m|$needle|s ) {
		# %sidebar  sb=rsidebar|t=More about the blog post%
		my $PID = trim($1);
		my $hashes = $2;
		my $description = trim($3);

		my @p = split(/ /, $PID);
		my$id = $p[0];
		my $Parms = "";
		if ($p[1]) {
			my $s = $PID;
			$s =~ s/^.*?\s//;
			$Parms = $s;
			#print "'$id' '$Parms'\n";
		}
		my $IDS = "";
		my $LinkStart = "";
		my $LinkEnd = "";
		my $NoHeading = "";
		my $Title = "";
		my $IDTitle = "";
		if ($Parms) {	# Do we have any parameters?
			#$Parms = _UnescapeSpecialChars($Parms);
			Parmify($Parms);
			#print "$id:\n"; PrintParms();
			if (exists($Parms{"noid"})) { $Parms{"ni"} = $Parms{"noid"} }
			if (exists($Parms{"nolink"})) { $Parms{"nl"} = $Parms{"nolink"} }
			if (exists($Parms{"noheading"})) { $Parms{"nh"} = $Parms{"noheading"} }
			if (exists($Parms{"nht"})) { $Parms{"nh"} = $Parms{"nht"} }
			if (exists($Parms{"ni"})) { $IDS = "#" }	# Suppress storing the ID
			if (exists($Parms{"nl"})) { $LinkStart = "<!-- "; $LinkEnd = " -->" }	# Ditto the link.
														# "Suppress storing" == "Comment out"
			if (exists($Parms{"nh"})) { $NoHeading = "1" }	# Suppress displaying the heading?
			if (exists($Parms{"ht"})) { $Parms{"t"} = $Parms{"ht"} }
			if (exists($Parms{"t"})) { 	# 'Hover text' that is displayed in the link.
				$Title = ' title="' . $Parms{"t"} . '"'; 
				$IDTitle = $Parms{"t"};
			}
		}
		
		my $level=length($hashes);
		my $MDescription = Markify($description);
		#print "'$id' '$Parms' '$IDS' '$hashes' '$level' '$MDescription'\n";
				
		$IDList .= "$IDS$id|$level|$description|$IDTitle\n";
		$NameLinks.= "  $LinkStart<li><a href=\"#$id\"$Title>$MDescription</a></li>$LinkEnd\n";
			
		my $chunk = "";
		$chunk .= "\n<div id=\"$id\">\n";
		$chunk .= "<a name=\"$id\"></a>\n";
		$chunk .= "</div>\n\n";
		unless ($NoHeading) { 
			# $chunk .= $hashes . $description . "\n" 
			$chunk .= "\n<h" . $level . $Title . ">" . $description . "</h$level>\n";
		}
		
		#print $chunk; exit 0;
		
		$s =~ s|$needle|$chunk|s;	# Out with the old, in with the new.
	}
	
	# End of PreFilter subroutine
	return $s;
}


sub Parmify {
	my $s = shift;
	
	%Parms = ();	# Empty out the %Parms hash before we start.
	$s = trim($s);
	my @l = split(/\|/, $s);
	my ($AParm, $key, $value);
	foreach $AParm (@l) {
		#print " $AParm\n";
		($key, $value) = split(/=/, $AParm);
		$key = lc(trim($key)); $value = trim($value);
		if ($key) { $Parms{$key} = $value; }
	}
	if (6 == 9) {		# Debugging....
		foreach $key (keys %Parms) {
			$value = $Parms{$key};
			print " '$key' = '$value'\n";
		}
	}
}

sub PrintParms {
	my ($key, $value);
	foreach $key (keys %Parms) {
		$value = $Parms{$key};
		print " '$key' = '$value'\n";
	}
}

# Changes <p>{whatever}</p> to <p class="$class">{whatever}</p>
# Well, actually changes <$tag> to <$tag class="$class">{whatever}</$tag>
# (But $tag defaults to 'p'.)
sub Classify {
	my $s = shift;
	my $class = shift;
	my $tag = shift;
	unless ($tag) { $tag = "p"; }
	
	my $mark = "--PonyTrots--";
	my $haystack = "<$tag>(.*?)</$tag>";
	while ( $s =~ m|$haystack|s ) {
		my $chunk = $1;
		my $needle = "<$mark class=\"$class\">" . $chunk . "</$mark>";
		$s =~ s|$haystack|$needle|s;	# Partial swapout
	}
	while ( $s =~ s|$mark|$tag|sgi ) { } # Finish the swapout
	return $s;
}

# Run a string through Markdown() and then clean it
sub Markify {
	my $s = shift;
	
	unless ($s) { return }
	$s = Markdown($s);
	$s =~ s|<p>||g;
	$s =~ s|</p>||g;
	$s =~ s|\n||g;
	return $s;
}

# Frogdown changes we make *after* running the text through Markdown().
# Which means that the text is now HTML.
sub PostFilter {
	my $s = shift;
	
	#$s =~ s|%af%|$AlternateFontBegin|sgi;	# Alternate font
	#$s =~ s|%/af%|$AlternateFontEnd|sgi;	# Say, san-serif if the normal font is a serif font.
	$s =~ s|%b%|<b>|sgi;			# Bold
	$s =~ s|%/b%|</b>|sgi;
	$s =~ s|%i%|<i>|sgi;		# Italic
	$s =~ s|%/i%|</i>|sgi;
	$s =~ s|%bi%|<b><i>|sg;		# BoldItalic (is a distinct font, dammit.)
	$s =~ s|%/bi%|</i></b>|sg;
#	if ($UseExtended) {	# $UseExtended is deprecated.
		$s =~ s|%em%|<em>|sg;		# Emphasize
		$s =~ s|%/em%|</em>|sg;
		$s =~ s|%strong%|<strong>|sg;	# Strong
		$s =~ s|%/strong%|</strong>|sg;	
#	}
	$s =~ s|%#%|<!-- |sg;		# Inline comment
	$s =~ s|%/#%| -->|sg;	
	$s =~ s|%tt%|<tt>|sgi;		# TT
	$s =~ s|%/tt%|</tt>|sgi;
	$s =~ s|%s%|<s>|sgi;		# Strikout
	$s =~ s|%/s%|</s>|sgi;
	#$s =~ s|\\ |&nbsp;|sg;		# Non break space
		
	#$s =~ s|\n?<p></blockquote></p>|</blockquote>|sgi;	# Remove extra <p></p> that Markdown()
														# Sometimes adds.
	$s =~ s|%blockquote%|\n<blockquote><p>|sgi;		# Block quote
	$s =~ s|%/blockquote%|</p></blockquote>\n|sgi;
														
	$s =~ s|%br%|<br>|sgi;		# Line break
	
	# Clears CSS floats.  Sometimes it is needed.
	$s =~ s|<p>%clearfloats%</p>|%clearfloats%|sgi;	# First, get rid of unwanted '<p>' tags.	
	$s =~ s|%ClearFloats%|<div style="clear\: both"></div>|sgi;
	
	my $WarningStart = "<span class=\"warning\">";
	my $WarningEnd = "</span>";
	$s =~s|%warnxing%|$WarningStart|sgi;
	$s =~s|%/warnxing%|$WarningEnd|sgi;
	
	my $ButtonStart = "<span class=\"button\">";
	my $ButtonEnd = "</span>";
	$s =~s|%butxton%|$ButtonStart|sgi;
	$s =~s|%/butxton%|$ButtonEnd|sgi;

	my $needle = "";


	# Rough equivalent of <img src=...
	# %img i=images/frog-logo-lg.png|mwp=100|t=100\% and proud of it.%
	$needle = "%imgxx (.*?)%";
	while ($s =~ m|$needle|s) {
		my $Parms = $1;
		my $chunk = "";
		#print "img: found '$Parms'\n";
		
		if ($Parms) {
			#print "%imagebox% parms: '$Parms'\n";
			Parmify($Parms);	# Fill %Parms hash with parameters
		} else {
			#print "No %imagebox% parms.\n";
		}

		my $Image = "";
		my $alt = "";
		my $title = "";
		my $LSpan = "";
		my $RSpan = "";
		my $MW = $MaxImageWidth;	# From config file
		if (exists($Parms{"i"})) { $Parms{"img"} =  $Parms{"i"} }
		if (exists($Parms{"img"})) {
			# <img src = "%%ImgFile%%">
			$Image = $Parms{"img"};
			$Image =~ s|</*em>|_|g;	# Markdown changes '_img_1234.hpg' to '<em>img</em>1234.jpg'
			#unless ( -f $Image ) { $ErrLog .= "Image file '$Image' not found.\n" }
		}
		if (exists($Parms{"mwp"})) {	# MaxImageWidth as per cent
			my $Percent = $Parms{"mwp"};
			$Percent =~ s/%//g;
			#print "Maximum image Width Percent: $Percent\n";
			if ($Percent) {		# '80' for 80%
				$MW = $MaxImageWidth / (100/$Percent);
				# print "Maximum image Width in Pixels: $MW\n";
			}
		}
		if (exists($Parms{"f"})) { $Parms{"ib"} =  $Parms{"f"} }
		if (exists($Parms{"ib"})) {
			$RSpan = "</span>";
			my $value = $Parms{"ib"};
			#if ( ($value =~ m/^(r)|(l)|(bq)$/i) && 1 ) {	# Doesn't work, so disabling.
			#if ( ($value =~ m/^(r)$/i) || ($value =~ m/^(l)$/i) || ($value =~ m/^(bq)$/i) ) {
			#	$value = lc($1) . "imagebox";
			#}
			if ($value =~ m/^(qb)$/i) { $value = "c" }
			if ($value =~ m/^(l)$/i) { $LSpan = "<span style=\"float: left; padding-right: 1em\">";
			} elsif ($value =~ m/^(r)$/i) { $LSpan = "<span style=\"float: right; padding-left: 1em\">"; 
			} elsif ($value =~ m/^(c)$/i) { 
				$LSpan = "</p><center><div style=\"display: block; margin-right: auto; margin-left: auto;\">";
				$RSpan = "</div></center><p>";	# Need to figure out how to do it without <center>.
			} else { $RSpan = "";
			}
		}		
		if (exists($Parms{"a"})) { $Parms{"c"} =  $Parms{"a"} }
		if (exists($Parms{"c"})) { $alt = ' alt="' . $Parms{"c"} . '"' }
		if (exists($Parms{"ht"})) { $Parms{"t"} = $Parms{"ht"} }
		if (exists($Parms{"t"})) { $alt .= ' title="' . $Parms{"t"} . '"' }
		
		if ($Image) { ProcessImage($Image, $MW); }
				
		my $chunk = "";
		$chunk = "$LSpan<img src=\"$Image\"$alt$title>$RSpan";
		$s =~ s|$needle|$chunk|s;	# Out with the old, in with the new.
	}	# imgxx
	
	$needle = "%(imagebox|img)\s*?(.*?)%";
	$s =~ s|<p>($needle)</p>|\1|sg;	 # Get rid of unwanted '<p>' tags that Markdown adds.
	while ( $s =~ m|$needle|s) {
		# %imagebox  img=images/ss-win-perl5lib-2.jpg|ib=rimagebox|c=Defining PERL5LIB in Windows%
		my $OForceImageCreation = $ForceImageCreation;
		my $WhichOne = trim($1);
		my $Parms = trim($2);
		$Parms =~ s|</*em>|_|g;	# !!! Markdown changes '_' to '<em>' pairs.  If there are pairs of '_'s/
		#print "Imagebox: found '$Parms'\n";
				if ($Parms) {
			#print "%imagebox% parms: '$Parms'\n";
			#Parmify($Parms);	# Fill %Parms hash with parameters
			Parmify_byref($Parms, \%Parms);	# Fill %Parms hash with parameters
		} else {
			$ErrLog .= "No %imagebox% parmameters.  Nothing to do.\n"; return;
		}	

		
		my $chunk = "";
		my %IMParms;	# Parameters to pass to ProcessImage()
		my $head = $ImageBoxBegin;	# Template definition from config file.
		my $Image ="";
		my $OImage = "";
		my $MW = $MaxImageWidth;	# From config file
		if ($MW < 100) { $MW = 100 }
	
		my $FileAddition = "";	# 'fa=_thum' $image = "foo.jpg" will become "foo_thum.jpg"
		if (exists($Parms{"fa"})) {
			$FileAddition = $Parms{"fa"};
			$IMParms{"fa"} = $FileAddition;	# Will be passing this to ProcessImaage();
		}
		if (exists($Parms{"i"})) { $Parms{"img"} =  $Parms{"i"} }
		if (exists($Parms{"img"})) {
			# <img src = "%%ImgFile%%">
			$Image = $Parms{"img"};
			$Image =~ s|</*em>|_|g;
			$OImage = FixImageExtension(AddFileAddition($Image, $FileAddition));
			#unless ( -f $Image ) { $ErrLog .= "Image file '$Image' not found.\n" }
			$head =~ s/%%ImgFile%%/$OImage/gi;
		}
		if (exists($Parms{"mwp"})) {	# MaxImageWidth as per cent
			my $Percent = $Parms{"mwp"};
			$Percent =~ s/%//g;
			#print "Maximum image Width Percent: $Percent\n";
			if ($Percent) {		# '80' for 80%
				$MW = $MaxImageWidth / (100/$Percent);
				# print "Maximum image Width in Pixels: $MW\n";
			} else {
				$MW = "";	# If 'mwp' is 0 or blank, then don't resize the image.
				#print "Maximum image Width Percent: '$Percent' \$MW: $MW\n";
				#print "Will not resize '$Image' ('$OImage')\n";
			}
		}
		if (exists($Parms{"f"})) { $Parms{"ib"} =  $Parms{"f"} }
		if (exists($Parms{"ht"})) { $Parms{"t"} = $Parms{"ht"} }
		# $ForceImageCreation 
		if (exists($Parms{"fic"})) {	# Force image creation for just this one image.
			my $Value = 1;
			my $c = $Parms{"fic"};
			if ($c eq "") { $c = "1" }
			if ($c =~ m/^([01])/) { $Value = $1 } 
			#print "$Parms\nfic: \$c = '$c'  \$Value = '$Value'\n";  
			if ($Value) {
				$ForceImageCreation = 1;
			}
		}		
		# End of ecumenical

		if ($WhichOne =~ m/imagebox/i) {
			if (exists($Parms{"ib"})) {
				# <div class = "%%Imagebox%%">
				my $value = $Parms{"ib"};
				#if ( ($value =~ m/^(r)|(l)|(bq)$/i) && 1 ) {	# Doesn't work, so disabling.
				if ( ($value =~ m/^(r)$/i) || ($value =~ m/^(l)$/i) || ($value =~ m/^(bq)$/i) ) {
					$value = lc($1) . "imagebox";	# 'r' is short for 'rimagebox.  etc.
				}
				$head =~ s/%%Imagebox%%/$value/gi;	
			} else {
				$head =~ s/%%Imagebox%%//gi;	
			}

			if ( 6 == 9 ) {	# Need to think about this more.
				if (exists($Parms{"a"})) { $Parms{"c"} =  $Parms{"a"} }
				unless( exists($Parms{"cc"}) ) { 
					if ( exists($TempVars{"captionclass"}) ) {  
						$Parms{"cc"} = $TempVars{"captionclass"};
						print "Swapping in CaptionClass from .cfg file: '" . $Parms{"cc"} . "'\n";
					}
				}

				if (exists($Parms{"lc"})) {
					my $cs = ""; my $ce="";
					if (exists($Parms{"c"}) ) {
					
					}
				}
			}
			# Simple case of just one kind of caption text.
			if (exists($Parms{"c"})) {
				# <span class="%%CaptionClass%%">%%CaptionText%%</span>
				$head =~ s/%%CaptionText%%/$Parms{"c"}/gi;	
			} else {
				$head =~ s/%%CaptionText%%//gi;
			}
			if (exists($Parms{"cc"})) {
				# <p class="%%CaptionClass%%"><img src = "%%ImgFile%%"><br>
				$head =~ s/%%CaptionClass%%/$Parms{"cc"}/gi;
			} else {
				# Assumed to be defined in .cfg file, so we don't want to overwrite that.
				#$head =~ s/%%CaptionClass%%/centcaption/gi;	# %%CaptionClass%% is assumed to be defined in .cfg (!)
			}
			if (exists($Parms{"t"})) { 	#  title="%%TitleText%%
				$head =~ s/%%TitleText%%/$Parms{"t"}/gi;
			} else {
				$head =~ s/%%TitleText%%//gi;
			}
			$head = ParseTempVars($head);	# Expand any '%%TempVar%% variables in the boilerplate
			#print "$head\n";
			$s =~ s|$needle|$head|s;	# Out with the old, in with the new.
		} elsif ($WhichOne =~ m/img/i) {
			$chunk = "--$Image--";
			my $alt = "";
			my $title = "";
			my $LSpan = "";
			my $RSpan = "";
			if (exists($Parms{"ib"})) {
				$RSpan = "</span>";
				my $value = $Parms{"ib"};
				#if ( ($value =~ m/^(r)|(l)|(bq)$/i) && 1 ) {	# Doesn't work, so disabling.
				#if ( ($value =~ m/^(r)$/i) || ($value =~ m/^(l)$/i) || ($value =~ m/^(bq)$/i) ) {
				#	$value = lc($1) . "imagebox";
				#}
				if ($value =~ m/^(qb)$/i) { $value = "c" }
				if ($value =~ m/^(l)$/i) { $LSpan = "<span style=\"float: left; padding-right: 1em\">";
				} elsif ($value =~ m/^(r)$/i) { $LSpan = "<span style=\"float: right; padding-left: 1em\">"; 
				} elsif ($value =~ m/^(c)$/i) { 
					$LSpan = "</p><center><div style=\"display: block; margin-right: auto; margin-left: auto;\">";
					$RSpan = "</div></center><p>";	# Need to figure out how to do it without <center>.
				} else { $RSpan = "";
				}
			}
			if (exists($Parms{"c"})) { $alt = ' alt="' . $Parms{"c"} . '"' }	
			if (exists($Parms{"t"})) { $title = ' title="' . $Parms{"t"} . '"' }	
			
			$chunk = "$LSpan<img src=\"$OImage\"$alt$title>$RSpan";
			$s =~ s|$needle|$chunk|s;	# Out with the old, in with the new.
			
		} # $WhichOne case
		
		if ($Image) { ProcessImage($Image, $MW, $FileAddition) }
		$ForceImageCreation = $OForceImageCreation;	
	}	# imagebox|img
	
	# Get rid of bogus paragraphs.  Like
	# <p><!--  Start of included content  --></p>
	my $needle = "\n<p><!--\s(.*?)\s--></p>\n";
	while ( $s =~ m|$needle|s ) {
		my $chunk = "\n<!-- " . $1 . " -->\n";	
		 #print "$chunk";
		$s =~ s|$needle|$chunk|s;	# Out with the old, in with the new.
	}

	
	# %id was moved to PreFilter()
	$needle = "<p>\s*%ixxd (.*?)%\s*</p>";
	my $NameCnt = 0;
	while ($s =~ m|$needle|s) {
		my $s1 = $1;
			
		my @parms = split (/\|/g, $s1);
		my $ParmCnt = scalar @parms;
		my $cnt=0;
		foreach (@parms) { $parms[$cnt] = trim($parms[$cnt]); $cnt++ }
		my $id = "";
		my $level = 0;
		my $description = "";
		if ($ParmCnt < 1) {
			$id = $NameCnt;
		} elsif ($ParmCnt < 2 ) {
			$id = $parms[0];
		} elsif ($ParmCnt < 3 ) {
			$id = $parms[0];
			$level = $parms[1];
		} elsif ($ParmCnt < 4 ) {
			$id = $parms[0];
			$level = $parms[1];
			$description = $parms[2];
		}
		$IDList .= "$id|$level|$description\n";
		$NameLinks.= "  <li><a href=\"#$id\">$description</a></li>\n";
		
		my $chunk = "";
		$chunk .= "\n<div id=\"$id\"></div>\n";
		$chunk .= "<a name=\"$id\" rel=\"external\"></a>\n";
		$s =~ s|$needle|$chunk|s;	# Out with the old, in with the new.
		$NameCnt++;
	}

	# <a href="mailto:someone@example.com?Subject=Hello%20again" target="_top">someone@example.com</a>
	$needle = "%mailto (.*?)%";
	while ($s =~ m|$needle|s) {
		my $s1 = $1;
			
		my @parms = split (/\|/g, $s1);
		my $ParmCnt = scalar @parms;
		my $cnt=0;
		foreach (@parms) { $parms[$cnt] = trim($parms[$cnt]); $cnt++ }

		# For now.
		my $link = "";
		my $subject = "";
		$link = $parms[0];	# someone@example.com
		if ($ParmCnt) { $subject = '?Subject=' . $parms[1] }
		$subject =~ s/ /%20/g;
		while ( $link =~ s/<.*?>/_/) {}	# Markdown converts '_' to <em></em> pairs.  :-(
		
		my $chunk = "<a href=\"mailto:$link$subject\">$link</a>";
		$s =~ s|$needle|$chunk|s;	# Out with the old, in with the new.
	}	
	
	# Keep this one last.  Because it is likely easy to match.
	$needle = "%l (.*?)%";
	while ($s =~ m|$needle|s) {
		my $s1 = $1;
			
		my @parms = split (/\|/g, $s1);
		my $ParmCnt = scalar @parms;
		my $cnt=0;
		foreach (@parms) { $parms[$cnt] = trim($parms[$cnt]); $cnt++ }
		
		#print "%l '$s1' $ParmCnt\n";;
		my $link = "";
		my $description = "";
		my $title = "";
		if ($ParmCnt < 2) {
			$link = $parms[0];
			$description = $link;
		} else {
			$link = $parms[1];
			$description = $parms[0];
		}
		if ($ParmCnt > 2) {
			$title = ' title="' . $parms[2] . '"';
		}
		while ( $link =~ s/<.*?>/_/) {}	# Markdown converts '_' to <em></em> pairs.  :-(
		
		my $chunk = "<a href=\"$link\"$title>$description</a>";
		$s =~ s|$needle|$chunk|s;	# Out with the old, in with the new.
	}	
	$s =~ s|(\n)?<p></p>|\1|;
	while ( $s =~ s|\n\n\n|\n\n|g ) {}
	
	$s = _UnescapeSpecialChars($s);
	return $s;
}	# PostFilter()

sub ProcessImage {
	my $Image = shift;
	my $MaxWidth = shift;
	my $FileAddition = shift;
	
	my $ImageMagickConfigL = $ImageMagickConfig;
	#print "'$Image' '$ForceImageCreation' '$CompressQuality'\n"; 
	my ($Width, $Height);
	
	if ($MaxWidth =~ m/x/i) {
		($Width, $Height) = split(/x/i, $MaxWidth);
	} else {
		$Width = $MaxWidth;
		$Height = $MaxWidth;
	}
	$Width = trim($Width);
	$Height = trim($Height);	
	# Filter input parms (somewhat)
	if ($Width < 1) { $Width = 1 }
	if ($Height < 1) { $Height = 1 }
	
	#print "Maximum image Width in Pixels: $MaxWidth\n";
	my ($q, $ArgDelimit);
	if ( $^O =~ /MSWin.*/i ) {
			$q = '"';	# For quoting paths on command lines
			$ArgDelimit = "^";
	} else {
			$q = '"';	# not sure for Linux.1
			$ArgDelimit = "\\";
	}
	
	my $size = "";
	if ($MaxWidth) {
		$size = "-resize " . $Width . "x" . $Height . $ArgDelimit . ">";	# '^' for Windows, '\' for Linux !!!
		# The '>' is so ImageMagick won't enlarge a smaller image--we only want to downsize.
	}
	if ($Image =~ m/Perl5/i || 1 ) {
		#print "\$Image: $Image  \$MaxWidth: $MaxWidth\n";
		
		# Maybe fixable?  
		my $JustFile = return_file($Image);
		#if ($JustFile =~ m/^_/) { $ErrLog .= "Can't process an image with a leading '_': $Image\n"; return }

		#$FileAddition
		
		my $OImage = FixImageExtension($Image);
		my $Outfile = $OutDir . $OImage;	# $OutDir is a global that is set at the beginning.
		if ($FileAddition ) { 
			#print "Adding '$FileAddition' to '$Outfile'\n"; 
			$Outfile = $OutDir . AddFileAddition($OImage, $FileAddition);
			#print "Added  '$FileAddition' to '$Outfile'\n"; 
		}
		my $OutPath = return_path($Outfile);
		unless ( -d $OutPath ) {
			#print "Output directory '$OutPath' does not exist\n"; exit -1;
			print "Making $OutPath\n";
			mkpath($OutPath, 1);
		}
		
		my $OutExists = "";
		if ( -f $Outfile ) { $OutExists = "1"; }
		unless ( -f $Image ) {	$ErrLog .= "Image '$Image' does not exist\n"; return }

		my $JustOFile = return_file($Outfile);	# Output file name with path removed.
		my $JustOBase = return_fname($JustOFile);	# Everything before the final "."
		my $JustOExt = return_fext($JustOFile);	# everything after the final "."
		
		if ( $ForceImageCreation || !($OutExists) || 0 ) {
			if ( $^O =~ /MSWin.*/i ) {
				$Image =~ s|/|\\|g;
				$Outfile =~ s|/|\\|g;
			}
			# -sharpen 0x0.40 -unsharp 0x0.40+1+0  Maybe? -unsharp 0x0.35+1+0

			if ($CompressQuality) {
				if ($JustOExt =~ m/jpg/i) {
					# http://www.imagemagick.org/script/command-line-options.php#compress
					# http://www.imagemagick.org/script/command-line-options.php#quality
					$ImageMagickConfigL .= " -compress JPEG -quality $CompressQuality";
				}
			}
			
			my $Parms = "convert $q$Image$q $size $ImageMagickConfigL $q$Outfile$q";
			#print "$Parms\n";
			my $result = `$Parms`;
			$WorkLog .= "$Parms\n";
			if ($result) { chomp ($result); $ErrLog .= "result\n"; }
			if ( -f $Outfile ) { 
				$Parms = "touch -c $q$Outfile$q";
				#print "$Parms\n";
				$result = `$Parms`; 
				if ($result) { chomp ($result); $ErrLog .= "result\n"; }
			}
		}
	}
}

sub AddFileAddition {
	my $Image = shift;
	my $Addition = shift;
	
	#my $Ofname = return_fname($Image);
	#my $Ofext = return_fext($Image);
	#$Outfile = $OutDir . $Ofname . $FileAddition . "." . $Ofext;
	
	return return_fname($Image) . $Addition . "." . return_fext($Image);
}

sub FixImageExtension {
	my $Image = shift;
	
	my $fname = return_fname($Image);
	my $fext = return_fext($Image);
	
	if ( $fext =~ m/$MakeTheseJPEGs/i ) {
		my $s = "Changing '$Image' to ";
		$Image = $fname . ".jpg"; 
		$s .= "'$Image'\n";
		#print $s;
		
	}
	return $Image;
}

sub RemoveTags {
	my $s = shift;
	while ($s =~ s/<.*?>//g) { }
	return $s;
}

sub ReadConfigFile {
	my $config = shift;
	
	unless (-f $config) { $ErrLog .= "Config file $config not found\n"; $fatal++; return; }
	open (FILE,$config) || die "Could not open configuration file '$config': $!"; 
	my $cfile = join('', <FILE>);	# Assumes that 'undef $/;' has *not* beeen done.
	close FILE;
	#print $cfile;
	$cfile =~ s{\r\n}{\n}g; # DOS to Unix
	$cfile =~ s{\r}{\n}g; 	# Mac to Unix
	print "Read config file $config\n";
	# Global variables
	if ($cfile =~ m/Start of Variables ###\n(.*)\n### End of Variables/s) {
		my $s = $1;
		ParseConfigVars($s);
		#ParseTempVars($s);
		#ParseConfigVars($s);
		#print $s;
	}

	if ($cfile =~ m/Start of CSS Variables ###\n(.*)\n### End of CSS Variables/s) {
		my $s = $1;
		#print $s;
		ParseConfigCSSVars($s);
	}

	if ($cfile =~ m/Start of Span Variables ###\n(.*)\n### End of Span Variables/s) {
		my $s = $1;
		#print $s;
		ParseConfigSpanVars($s);
	}
		
	ParseHeaderVars($cfile);
}

sub ParseHeaderVars {
	my $cfile = shift;
	
	if ($cfile =~ m/Start of Header Begin Description ###\n(.*)### End of Header Begin Description/s) {
		$PageHeader = $1 . "<!-- End of header -->\n";
		#print $PageHeader;
	}
	if ($cfile =~ m/Start of Footer Begin Description ###\n(.*)### End of Footer Begin Description/s) {
		$PageFooter = "<!-- Beginning of footer -->\n" . $1;
		#print $PageFooter;
	}
	if ($cfile =~ m/Start of Sidebar Begin Description ###\n(.*)### End of Sidebar Begin Description/s) {
		$SideBarBegin ="\n<!-- Start of sidebar -->\n" . $1;
		#print $SideBarBegin;
	}
	if ($cfile =~ m/Start of Sidebar End Description ###\n(.*)### End of Sidebar End Description/s) {
		$SideBarEnd = $1 . "<!-- End of of sidebar -->\n";
		#print $SideBarEnd;
	}
	if ($cfile =~ m/Start of ImageBox Begin Description ###\n(.*)### End of ImageBox Begin Description/s) {
		$ImageBoxBegin = $1;
		#print $SideBarBegin;
	}
	#if ($cfile =~ m/Start of ImageBox End Description ###\n(.*)### End of ImageBox End Description/s) {
	#	$ImageBoxEnd = $1;
	#	#print $SideBarEnd;
	#}
		
}

sub ParseConfigSpanVars {
	my $s = shift;
	#print $s;
	my ($name, $value);
	my @v = split(/\n/g, $s);
CSS: foreach (@v) {
		# ignore lines starting with "#"
		if ( ! /^\s*#/ ) {
			chomp;
			# skip blank lines
			next if ( /^\s*$/ );
			#print "$_ \n";
			my $warlock = "~%~warlock~%~";
			s/\\=/$warlock/g;
			# split the line into variable and value
			($name, $value)=split("=");
			$name = trim($name);
			$value = trim($value);
			$value =~ s/$warlock/=/g;
			#print "'$name' = '$value'\n";
			# make the variable lowercase
			$name = lc($name);
			if ( $name =~ m/^_/) {
				$name =~ s/^_//;
				$SpanVars{$name} = ($value); 
				#$WorkLog .= "Found Span '_' template variable '$name' '$value'\n";
				#print "Found Span '_' template variable '$name' '$value'\n";
				next CSS;
			}
		}
	}
}

sub ParseConfigCSSVars {
	my $s = shift;
	#print $s;
	my ($name, $value);
	my @v = split(/\n/g, $s);
CSS: foreach (@v) {
		# ignore lines starting with "#"
		if ( ! /^\s*#/ ) {
			chomp;
			# skip blank lines
			next if ( /^\s*$/ );
			#print "$_  ";

			# split the line into variable and value
			($name, $value)=split("=");
			$name = trim($name);
			$value = trim($value);
			# make the variable lowercase
			$name = lc($name);
			if ( $name =~ m/^_/ ) {
				$name =~ s/^_//;
				$CSSVars{$name} = ($value); 
				#$WorkLog .= "Found CSS '_' template variable '$name' '$value'\n";
				next CSS;
			}
		}
	}
}

sub ParseConfigVars {
	my $s = shift;

	#print $s;
	#my ($s1, $s2, $s3, $s4, $s5);
	my ($name, $value);
	my @v = split(/\n/g, $s);
PCV:	foreach (@v) {
		# ignore lines starting with "#"
		if ( ! /^\s*#/ ) {
			chomp;

			# skip blank lines
			next if ( /^\s*$/ );
			#print "$_  ";

			# split the line into variable and value
			($name, $value)=split("=");
			# get rid of leading and trailing spaces in value and name
			$name =~ s/^\s+//;
			$name =~ s/\s+$//;
			$value =~ s/^\s+//;
			$value =~ s/\s+$//;
			# make the variable lowercase
			$name = lc($name);		

			#$WorkLog .= "Found '%%' template variable '$name' '$value'\n";
			if ( $name =~ m/^_/ ) {
				$name =~ s/^_//;
				$TempVars{$name} = ($value); 
				#$WorkLog .= "Found '_' template variable '$name' '$value'\n";
				next PCV;
			} elsif ( 0 ) {	# $name =~ m/^%%.*?%%$/ 
				#$name =~ s/^%%//;
				#$name =~ s/%%$//;
				#$TempVars{$name} = ($value); 
				#$ErrLog .= "Found deprecated '%%' template variable '$name' '$value' in config file.\n";
				$ErrLog .= "Should not happen.  Problem in ParseConfigVars()\n";
				#print "Found '%%' template variable '$name' '$value'\n";
				next PCV;
			}
			
			# Set global variables (that must already be defined globally!)
			if ($name eq "baseindir") { 
				$BaseInDir = ($value);
				$BaseInDir =~ s|\\|/|g;
				$BaseInDir = FixLinuxPath($BaseInDir);
				$BaseInDir = add_rslash($BaseInDir);
				if ($CurrentDrive) { $BaseInDir = ChangeDrive($BaseInDir, $CurrentDrive) }
				$JustDir = $pwd;
				$JustDir =~ s/$BaseInDir//i;
				$InDir = add_rslash($BaseInDir . $JustDir);
				#$JustDir = "";	# We are flatlanders
			} elsif ($name eq "baseoutdir") { 
				$BaseOutDir = ($value);
				$BaseOutDir =~ s|\\|/|g;
				$BaseOutDir = FixLinuxPath($BaseOutDir);
				$BaseOutDir = add_rslash($BaseOutDir);
				if ($CurrentDrive) { $BaseOutDir = ChangeDrive($BaseOutDir, $CurrentDrive) }
				$OutDir = add_rslash($BaseOutDir . $JustDir);
			}
			
			if ($name =~ m/Include/i) { 
				my $FileName = ($value);
				my $s = "";
				my $FullName = "";
				if ( -f $FileName ) {	# Current directory? 
					$FullName = $FileName;
				} elsif ( -f $ConfigPath . $FileName ) { # Config directory?
					$FullName = $ConfigPath . $FileName;
				} else {
					$ErrLog .= "Include file '$FileName' not found\n";
					#print "Include file '$FileName' not found\n"; exit 0;
				}
				# Doesn't work(!)
				#unless ( readfile_byref($Fullname, \$s) ) { print "Include file '$FileName' not found\n" }
				if ($FullName) { 
					my $s = readfile($FullName);
					print "Read config file $FullName\n"; 
					ParseConfigVars($s);
					#print "$s"; exit 0;
				}			
			}
			
			if ($name eq "file") { $FilesToBeCopied .= ($value)  . "\n"; }
			if ($name eq "ismobile") { $IsMobile = ($value); }
			if ($name eq "imagemagickconfig") { $ImageMagickConfig = ($value) }
			if ($name eq "maximagewidth") { $MaxImageWidth = ($value) }
			if ($name =~ m/TemplateVarsInBody/i) { $TemplateVarsInBody = ($value) }
			if ($name =~ m/UseExtended/i) { $UseExtended = ($value) }

			if ($name =~ m/MakeTheseJPEGs/i) { $MakeTheseJPEGs = ($value) }
			if ($name =~ m/MakeHTMLFile/i) { $MakeHTMLFile = ($value) }
			if ($name =~ m/MakeTextFile/i) { $MakeTextFile = ($value) }
			if ($name =~ m/CreateDirs/i) { $CreateDirs = ($value) }
			if ($name =~ m/UseSmartyPants/i) { $UseSmartyPants = ($value) }
			if ($name =~ m/ForceImageCreation/i) { $ForceImageCreation = ($value) }
			if ($name =~ m/CompressQuality/i) { $CompressQuality = ($value) }
			if ($name =~ m/Debug/i) { $Debug = ($value) }
			if ($name =~ m/ShowVars/i) { $ShowVars  = ($value) }
			if ($name =~ m/OnlyPrintConfigs/i) { 
				$OnlyPrintConfigs = ($value);
				if ($DDebug && 0) { print "\$OnlyPrintConfigs: $value\n"; }
			}
			if ($name =~ m/MakeTextFile/i) { $MakeTextFile = ($value) }
			#if ($name =~ m//i) { $ = ($value) }
			#if ($name =~ m//i) { $ = ($value) }	
			#if ($name =~ m//i) { $ = ($value) }
		}	# if ( ! /^#/ ) {
	}	# foreach (@v) {
}

# Perl doesn't expand '~' into $ENV{'HOME'  So...
sub FixLinuxPath {
	my $s = shift;
	
	if ( $s =~ m/~/ ) {
		if ($DDebug) { 
			print "Changing '~' to '$HomePath' in '$s'\n";
		}
		if ($s =~ m|/var/chroot/|i ) { print "'~' expansion will not work on this machine.\n"; exit -1; }
		$s =~ s|~|$HomePath|;	# $HomePath is a global and we filled this at the beginning.
		$s =~ s|//|/|g;			# Value?  It all depends--search for $HomePath from the top of the script.
		if ($DDebug) { print "  is now '$s'\n" }
	}							
	return $s
}

# Pick off configuration variables from beginning of .md file
sub GetHeadVars {
	my $cfile = shift;
	
	# Get all the regular variables so they can be parsed
	if ($cfile =~ m/Start of Variables ###\n(.*)\n### End of Variables/s) {
		#print "Found variables at the beginning of the file\n";
		my $s = $1;
		ParseConfigVars($s);
		# Now delete the variables from the beginning of the file
		#$cfile =~ s/.*?### End of Variables ###\n//s;
		#$cfile = "<!-- Removed variables from the beginning of the file -->\n" . $cfile;
		# Now cover up the variables (as HTML comments)
		if ($Debug && 0) {
			# Surround the variables with HTML comment tags
			$cfile = "<!-- \n" . $cfile;
			$cfile =~ s/(.*?### End of Variables ###\n)/\1 -->\n\n/s;
		} else {
			# Just get rid of them
			$cfile =~ s/(### Start of Variables ###\n.*?### End of Variables ###\n)//s;
		}
	}
		# Get all the CSS variables so they can be parsed
	if ($cfile =~ m/Start of CSS Variables ###\n(.*)\n### End of CSS Variables/s) {
		print "Found CSS variables at the beginning of the file\n";
		my $s = $1;
		ParseConfigCSSVars($s);
		# Now delete the variables from the beginning of the file
		#$cfile =~ s/.*?### End of Variables ###\n//s;
		#$cfile = "<!-- Removed variables from the beginning of the file -->\n" . $cfile;
		# Now cover up the variables (as HTML comments)
		if ($Debug && 0) {
			# Surround the variables with HTML comment tags
			$cfile = "<!-- \n" . $cfile;
			$cfile =~ s/(.*?### End of CSS Variables ###\n)/\1 -->\n\n/s;
		} else {
			# Just get rid of them
			$cfile =~ s/(### Start of CSS Variables ###\n.*?### End of CSS Variables ###\n)//s;
		}
	}
		# Get all the CSS variables so they can be parsed
	if ($cfile =~ m/Start of Span Variables ###\n(.*)\n### End of Span Variables/s) {
		#print "Found span variables at the beginning of the file\n";
		my $s = $1;
		ParseConfigSpanVars($s);
		# Now delete the variables from the beginning of the file
		#$cfile =~ s/.*?### End of Variables ###\n//s;
		#$cfile = "<!-- Removed variables from the beginning of the file -->\n" . $cfile;
		# Now cover up the variables (as HTML comments)
		if ($Debug && 0) {
			# Surround the variables with HTML comment tags
			$cfile = "<!-- \n" . $cfile;
			$cfile =~ s/(.*?### End of Span Variables ###\n)/\1 -->\n\n/s;
		} else {
			# Just get rid of them
			$cfile =~ s/(### Start of Span Variables ###\n.*?### End of Span Variables ###\n)//s;
		}
	}
	return $cfile;
}


sub ParseSpanVars {
	my $s = shift;
	#print "Parsing span variables\n";
	
	my ($key, $value);
	foreach $key (keys %SpanVars) {
		$value = $SpanVars{$key};
		#print "ParseSpanVars: Key: '$key' Value: '$value'\n"; 
		my $needle = "%$key%(.*?)%/$key%";
		while ( $s =~ m/$needle/s ) {
			my $chunk = $value . $1 . "</span>";
			$s =~ s|$needle|$chunk|s;
		}
		#if ( $s =~ s/%$key%/$value/gi ) {
			#print "  Matched template variable: Key: '$key' Value: '$value'\n"; 
		#}
	}
	return $s;
}

sub ParseTempVars {
	my $s = shift;
	#print "Parsing template variables\n";
	#return $s;	
	my ($key, $value);
	foreach $key (keys %TempVars) {
		$value = $TempVars{$key};
		#print "ParseTempVars: Key: '$key' Value: '$value'\n"; 
		if ( $s =~ s/%%$key%%/$value/gi ) {
			#print "  Matched template variable: Key: '$key' Value: '$value'\n"; 
		}
	}
	return $s;
}

sub ParseCSSVars {
	my $s = shift;
	
	my ($key, $value);
	foreach $key (keys %CSSVars) {
		$value = $CSSVars{$key};
		#print "ParseCSSVars: Key: '$key' Value: '$value'\n"; 
		if ( $s =~ s/%%$key%%/$value/gi ) {
			#print "  Matched template variable: Key: '$key' Value: '$value'\n"; 
		}
	}
	return $s;
}

sub addhead {
	#my $s = shift;	
	my $x = "";
	if ($PageHeader) { 
		$PageHeader = ParseTempVars($PageHeader);
		#print $PageHeader;
		return $PageHeader; 
	} else {
		$x .= "<html><head><title>" . $BaseName . "</title>\n";
		$x .= "</head>\n";
		$x .= "<body>\n";
		$x .= "<!-- End of heading -->\n";

		return $x;
	}
}

sub addtail {

	#my $s = shift;
	my $x = "";
	if ($PageFooter) { 
		$PageFooter = ParseTempVars($PageFooter);
		return $PageFooter;
	} else {
		$x .= "<!-- Beginning of tail -->\n";
		$x .= "</body></html>\n";	
		return $x;
	}
}

# Word wrap a string.
sub wrapit {
	my $comments = shift;
	my $NumCols = shift;
	if ( ! defined $NumCols ) { $NumCols = "79"; }
	if (defined $comments) {
		my $delim = "\n";
		
		$comments = $comments . $delim;
		my @pars = split (/$delim/, $comments);
		my ($s, $s1, $par);
		# Loop through pars, collect them all in $s.
		$s = "";
		foreach $par (@pars) {
			#$par = trim($par);
			$Text::Wrap::columns=$NumCols;
			$Text::Wrap::separator="\n";
			$par = wrap("","",$par);
			#print "OnePar: |$par|\n";
			$s .= $par . $delim;
		}
	$s =~ s/\n+$//s;	# Remove trailing '\n's
	return $s . "";
	} else {
		return "";
	}
}

# Cribbed from Markdown.pl
sub _EncodeBackslashEscapes {
#
#   Parameter:  String.
#   Returns:    The string, with after processing the following backslash
#               escape sequences.
#
    local $_ = shift;

	if ( 6 == 9 ) {
		s! \\\\  !$g_escape_table{'\\'}!gx;		# Must process escaped backslashes first.
		s! \\`   !$g_escape_table{'`'}!gx;
		s! \\\*  !$g_escape_table{'*'}!gx;
		s! \\_   !$g_escape_table{'_'}!gx;
		s! \\\{  !$g_escape_table{'{'}!gx;
		s! \\\}  !$g_escape_table{'}'}!gx;
		s! \\\[  !$g_escape_table{'['}!gx;
		s! \\\]  !$g_escape_table{']'}!gx;
		s! \\\(  !$g_escape_table{'('}!gx;
		s! \\\)  !$g_escape_table{')'}!gx;
		s! \\>   !$g_escape_table{'>'}!gx;
		s! \\\#  !$g_escape_table{'#'}!gx;
		s! \\\+  !$g_escape_table{'+'}!gx;
		s! \\\-  !$g_escape_table{'-'}!gx;
		s! \\\.  !$g_escape_table{'.'}!gx;
		s{ \\!  }{$g_escape_table{'!'}}gx;
	}
    s{ \\%  }{$g_escape_table{'%'}}gx;
	#s/\\ /{$g_escape_table{' '}/g;
	#my $s = "### non break space ###";
	#s/\\$s/g;
    s{ \\\|  }{$g_escape_table{'|'}}gx;
	s|\\=|$g_escape_table{'='}|g;

	
	my $s = "!### Non break space ###!";
	s|\\\ |$s|g;
	
    return $_;
}

# Cribbed from Markdown.pl
sub _UnescapeSpecialChars {
#
# Swap back in all the special characters we've hidden.
#
	my $text = shift;
	#return $text; 
	
	while( my($char, $hash) = each(%g_escape_table) ) {
		$text =~ s/$hash/$char/g;
	}

	my $s = "!### Non break space ###!";
	$text =~ s/$s/&nbsp;/g;
	
    return $text;
}
